# 点
from .Point import Point
from .Points import Points

# 线
from .Polyline import Polyline
from .Circle import Circle

# 面
from .RectFace import RectFace

from .RPolygonFace import RPolygonFace
from .HollowRPolygonFace import HollowRPolygonFace

from .CircleFace import CircleFace
from .TorusFace import TorusFace
from .EllipseFace import EllipseFace

# 体

# 多面体
from .Box import Box

from .Prism import Prism
from .PrismSide import PrismSide

from .Pyramid import Pyramid
from .PyramidSide import PyramidSide

from .PyramidTrustum import PyramidTrustum
from .PyramidTrustumSide import PyramidTrustumSide

# 曲面体
from .Cylinder import Cylinder
from .CylinderSide import CylinderSide

from .Cone import Cone
from .ConeSide import ConeSide
from .ConeTrustum import ConeTrustum
from .ConeTrustumSide import ConeTrustumSide

from .Sphere import Sphere
from .Icosphere import Icosphere
from .SphericalCap import SphericalCap
from .SphericalCapTop import SphericalCapTop

# 正多面体
from .Tetrahedron import Tetrahedron # 正四面体
from .Hexahedron import Hexahedron # 正六面体
from .Octahedron import Octahedron # 正八面体
from .Dodecahedron import Dodecahedron # 正十二面体
from .Icosahedron import Icosahedron # 正二十面体

# 通用函数曲面
from .Surf import Surf
from .FSurf import FSurf
from .CylindricalFSurf import CylindricalFSurf
from .SphericalFSurf import SphericalFSurf

# 变形体
from .Rotator import Rotator
from .Extruder import Extruder

# 特殊几何体
from .Floor import Floor
from .Torus import Torus
from .TrefoilKnot import TrefoilKnot
from .CoordSys import CoordSys
from .ImageQuad import ImageQuad


from ..Mesh import Mesh
from ..algorithm import cos_angle_of

from glass import Vertex

import math
import glm
import numpy as np

class Rotator(Mesh):
    def __init__(self, section:(list,tuple,np.ndarray),
                 axis_start:glm.vec3=glm.vec3(0), axis_stop:glm.vec3=glm.vec3(0,0,1),
                 n_divide:int=100, start_angle:float=0, span_angle:float=360,
                 n_lat_divide:int=100,
                 color:(glm.vec3,glm.vec4)=glm.vec4(0.396, 0.74151, 0.69102, 1), back_color:(glm.vec3,glm.vec4)=None,
                 normalize_tex_coord:bool=False,
                 name:str="", block=True, surf_type:Mesh.SurfType=Mesh.SurfType.Auto):
        Mesh.__init__(self, color=color, back_color=back_color, name=name, block=block, surf_type=surf_type)
        self.__section = section
        self.__axis_start = axis_start
        self.__axis_stop = axis_stop
        self.__start_angle = start_angle
        self.__span_angle = span_angle
        self.__n_divide = n_divide
        self.__n_lat_divide = n_lat_divide
        self.__normalize_tex_coord = normalize_tex_coord
        self.start_building()

    def build(self):
        vertices = self.vertices
        indices = self.indices
        section = self.__used_section
        axis_start = self.__axis_start
        axis_stop = self.__axis_stop
        start_angle = self.__start_angle/180*math.pi
        span_angle = self.__span_angle/180*math.pi
        n_divide = self.__n_divide
        normalize_tex_coord = self.__normalize_tex_coord

        i_vertex = 0
        i_index = 0

        # 计算原点到轴的垂足 H
        A = axis_start
        B = axis_stop
        C = glm.vec3(0)
        AC = C-A
        AB = B-A
        AB = glm.normalize(AB)
        H = A + glm.dot(AC, AB) * AB

        # 计算 section 长度 L，section 距轴的最长距离 D
        L, D = 0, 0
        len_AB = glm.length(AB)
        len_section = len(section)
        for i in range(len_section):
            if i > 0:
                L += glm.length(section[i] - section[i-1])

            d = glm.length(glm.cross(AB, section[i]-A))/len_AB
            if d > D:
                D = d

        for i in range(n_divide):
            theta = start_angle + span_angle*i/(n_divide - 1)
            half_theta = theta / 2
            quat = glm.quat(math.cos(half_theta), math.sin(half_theta)*AB)

            l = 0
            for j in range(len_section):
                pos = quat * (section[j] - H) + H
                vertex = Vertex()
                vertex.position = pos
                vertex.normal = glm.vec3(0)

                if j > 0:
                    l += glm.length(section[j] - section[j-1])

                if not normalize_tex_coord:
                    vertex.tex_coord = glm.vec3(D*theta, 1-l, 0)
                else:
                    vertex.tex_coord = glm.vec3(D/L*theta, 1-l/L, 0)

                vertices[i_vertex] = vertex
                i_vertex += 1

                if i > 0 and j > 0:
                    triangle = glm.uvec3(0, 0, 0)
                    triangle[0] = i_vertex - 1
                    triangle[1] = i_vertex - 2
                    triangle[2] = i_vertex - 2 - len_section
                    indices[i_index] = triangle
                    i_index += 1
                    self.generate_temp_TBN(vertices[triangle[0]], vertices[triangle[1]], vertices[triangle[2]])

                    triangle = glm.uvec3(0, 0, 0)
                    triangle[0] = i_vertex - 1
                    triangle[1] = i_vertex - 2 - len_section
                    triangle[2] = i_vertex - 1 - len_section
                    indices[i_index] = triangle
                    i_index += 1
                    self.generate_temp_TBN(vertices[triangle[0]], vertices[triangle[1]], vertices[triangle[2]])

                    yield

        del vertices[i_vertex:]
        del indices[i_index:]

    @property
    def section(self):
        return self.__section
    
    @section.setter
    @Mesh.param_setter
    def section(self, section:(list,tuple,np.ndarray)):
        self.__section = section

    @property
    def __used_section(self):
        should_duplicate_indices = set()
        for i in range(1, len(self.__section)-1):
            v1 = self.__section[i-1] - self.__section[i]
            v2 = self.__section[i+1] - self.__section[i]
            cos_angle = cos_angle_of(v1, v2)
            if cos_angle > -0.9:
                should_duplicate_indices.add(i)

        used_section = []
        for i in range(len(self.__section)):
            used_section.append(self.__section[i])
            if i in should_duplicate_indices:
                used_section.append(self.__section[i])

        return used_section
    
    @property
    def axis_start(self):
        return self.__axis_start
    
    @axis_start.setter
    @Mesh.param_setter
    def axis_start(self, axis_start:glm.vec3):
        self.__axis_start = axis_start

    @property
    def axis_stop(self):
        return self.__axis_stop
    
    @axis_stop.setter
    @Mesh.param_setter
    def axis_stop(self, axis_stop:glm.vec3):
        self.__axis_stop = axis_stop
    
    @property
    def n_divide(self):
        return self.__n_divide
    
    @n_divide.setter
    @Mesh.param_setter
    def n_divide(self, n_divide:int):
        self.__n_divide = n_divide
    
    @property
    def start_angle(self):
        return self.__start_angle
    
    @start_angle.setter
    @Mesh.param_setter
    def start_angle(self, start_angle:float):
        self.__start_angle = start_angle
    
    @property
    def span_angle(self):
        return self.__span_angle
    
    @span_angle.setter
    @Mesh.param_setter
    def span_angle(self, span_angle:float):
        self.__span_angle = span_angle
    
    @property
    def normalize_tex_coord(self):
        return self.__normalize_tex_coord
    
    @normalize_tex_coord.setter
    @Mesh.param_setter
    def normalize_tex_coord(self, flag:bool):
        self.__normalize_tex_coord = flag



from .ShaderEffect import ShaderEffect
from .PostProcessEffect import PostProcessEffect
from .GaussBlur import GaussBlur
from .MulFilter import MulFilter
from glass import sampler2D
from OpenGL import GL
import os

class SSAOVisibility(ShaderEffect):

    def __init__(self):
        self_folder = os.path.dirname(os.path.abspath(__file__))
        ShaderEffect.__init__(self, self_folder + "/../glsl/PostProcessEffects/SSAO_visibility.glsl", GL.GL_R32F)

class SSAOEffect(PostProcessEffect):
        
    def __init__(self):
        PostProcessEffect.__init__(self)

        self._ssao_visibility = SSAOVisibility()
        self._gauss_blur = GaussBlur(kernel_shape=16, internal_format=GL.GL_R32F)
        self._mul_filter = MulFilter()

        self.radius:float = 0.2
        self.samples:int = 64
        self.power:float = 2.2

    def apply(self, screen_image:sampler2D)->sampler2D:
        self._ssao_visibility.camera = self.camera
        self._ssao_visibility.view_pos_map = self.view_pos_map
        self._ssao_visibility.view_normal_map = self.view_normal_map
        self._ssao_visibility.depth_map = self.depth_map
        self._ssao_visibility["radius"] = self.radius
        self._ssao_visibility["samples"] = self.samples
        self._ssao_visibility["power"] = self.power

        ssao_image = self._ssao_visibility.apply(screen_image)
        ssao_image = self._gauss_blur.apply(ssao_image)

        self._mul_filter["filter_image"] = ssao_image
        self._mul_filter["filter_channels"] = 1
        return self._mul_filter.apply(screen_image)
    
    def need_pos_info(self) -> bool:
        return True

    def draw_to_active(self, screen_image:sampler2D):
        self._ssao_visibility.camera = self.camera
        self._ssao_visibility.view_pos_map = self.view_pos_map
        self._ssao_visibility.view_normal_map = self.view_normal_map
        self._ssao_visibility.depth_map = self.depth_map
        self._ssao_visibility["radius"] = self.radius
        self._ssao_visibility["samples"] = self.samples
        self._ssao_visibility["power"] = self.power

        ssao_image = self._ssao_visibility.apply(screen_image)
        ssao_image = self._gauss_blur.apply(ssao_image)

        self._mul_filter["filter_image"] = ssao_image
        self._mul_filter["filter_channels"] = 1
        self._mul_filter.draw_to_active(screen_image)
    


from .Icosahedron import Icosahedron

from ..Mesh import Mesh

from glass.utils import checktype
from glass import Vertex

import glm
import math
import copy

class Icosphere(Mesh):

    __base_vertices = []
    __base_indices = []

    @checktype
    def __init__(self, radius:float=1,
                 color:(glm.vec3,glm.vec4)=glm.vec4(0.396, 0.74151, 0.69102, 1), back_color:(glm.vec3,glm.vec4)=None,
                 n_levels:int=4, name:str="", block=True):
        Mesh.__init__(self, color=color, back_color=back_color, name=name, block=block)
        self.__radius = radius
        self.__n_levels = n_levels
        self.start_building()

    def __generate_base_data(self):
        if Icosphere.__base_indices:
            return

        Icosphere.__base_indices = copy.deepcopy(Icosahedron.base_indices)
        Icosphere.__base_vertices = []

        half_theta = math.atan(2)/4
        cos_half_theta = math.cos(half_theta)
        sin_half_theta = math.sin(half_theta)
        q = glm.quat(cos_half_theta, 0, sin_half_theta, 0)
        for base_pos in Icosahedron.base_positions:
            pos = q * base_pos
            if glm.length(pos - glm.vec3(0,0,1)) < 1E-6:
                pos = glm.vec3(0,0,1)
            elif glm.length(pos - glm.vec3(0,0,-1)) < 1E-6:
                pos = glm.vec3(0,0,-1)
            Icosphere.__base_vertices.append(Icosphere.__create_vertex(pos))

        for index in Icosphere.__base_indices:
            vertex0 = Icosphere.__base_vertices[index[0]]
            vertex1 = Icosphere.__base_vertices[index[1]]
            vertex2 = Icosphere.__base_vertices[index[2]]

            v1 = vertex1.position - vertex0.position
            v2 = vertex2.position - vertex0.position
            normal = glm.cross(v1, v2)

            for i in range(3):
                vertex = Icosphere.__base_vertices[index[i]]
                if vertex.tex_coord.t == 0 or vertex.tex_coord.t == 1:
                    new_vertex = Icosphere.__create_vertex(vertex.position, normal)
                    new_index = len(Icosphere.__base_vertices)
                    Icosphere.__base_vertices.append(new_vertex)
                    index[i] = new_index

    def build(self):
        self.is_closed = True
        vertices = self.vertices
        indices = self.indices
        self.__generate_base_data()
  
        radius = self.__radius
        n_levels = self.__n_levels

        _indices = Icosphere.__base_indices

        i_vertex = 0
        i_index = 0
        for vertex in Icosphere.__base_vertices:
            new_vertex = copy.deepcopy(vertex)
            new_vertex.position = radius*new_vertex.position
            new_vertex.tangent = radius*new_vertex.tangent
            new_vertex.bitangent = radius*new_vertex.bitangent
            vertices[i_vertex] = new_vertex
            i_vertex += 1

        for _ in range(n_levels):
            i_index = 0
            for index in _indices:
                vertex0 = vertices[index[0]]
                vertex1 = vertices[index[1]]
                vertex2 = vertices[index[2]]

                new_pos0 = glm.normalize(vertex1.position + vertex2.position)
                new_pos1 = glm.normalize(vertex0.position + vertex2.position)
                new_pos2 = glm.normalize(vertex0.position + vertex1.position)

                new_vertex0 = Icosphere.__create_vertex(new_pos0, radius=radius)
                new_vertex1 = Icosphere.__create_vertex(new_pos1, radius=radius)
                new_vertex2 = Icosphere.__create_vertex(new_pos2, radius=radius)                

                start_index = i_vertex
                vertices[i_vertex] = new_vertex0
                i_vertex += 1

                vertices[i_vertex] = new_vertex1
                i_vertex += 1

                vertices[i_vertex] = new_vertex2
                i_vertex += 1

                indices[i_index] = glm.uvec3(index[0], start_index+2, start_index+1)
                i_index += 1

                indices[i_index] = glm.uvec3(index[1], start_index  , start_index+2)
                i_index += 1

                indices[i_index] = glm.uvec3(index[2], start_index+1, start_index  )
                i_index += 1

                indices[i_index] = glm.uvec3(start_index, start_index+1, start_index+2)
                i_index += 1

                yield

            _indices = list(indices)
        
        del vertices[i_vertex:]
        del indices[i_index:]

        for index in indices:
            vertex0 = vertices[index[0]]
            vertex1 = vertices[index[1]]
            vertex2 = vertices[index[2]]
            s0 = vertex0.tex_coord.s
            s1 = vertex1.tex_coord.s
            s2 = vertex2.tex_coord.s

            if abs(s0 - s1) > 0.5 and abs(s0 - s2) > 0.5:
                new_vertex = copy.deepcopy(vertex0)
                new_vertex.tex_coord.s += (1 if s1 + s2 > 1 else -1)
                new_index = len(vertices)
                vertices.append(new_vertex)
                index[0] = new_index
                
            if abs(s1 - s0) > 0.5 and abs(s1 - s2) > 0.5:
                new_vertex = copy.deepcopy(vertex1)
                new_vertex.tex_coord.s += (1 if s0 + s2 > 1 else -1)
                new_index = len(vertices)
                vertices.append(new_vertex)
                index[1] = new_index
                
            if abs(s2 - s0) > 0.5 and abs(s2 - s1) > 0.5:
                new_vertex = copy.deepcopy(vertex2)
                new_vertex.tex_coord.s += (1 if s0 + s1 > 1 else -1)
                new_index = len(vertices)
                vertices.append(new_vertex)
                index[2] = new_index

            yield

    def __get_info(position, normal=None):
        dem = math.sqrt(position.x*position.x + position.y*position.y)
        if dem == 0:
            s = 0
            tangent = glm.vec3(0, 2*math.pi, 0)
            bitangent = glm.vec3(-math.pi, 0, 0)
            if normal is not None:
                theta = math.atan2(normal.y, normal.x)
                phi = math.pi/2 if normal.z > 0 else -math.pi/2
                sin_theta = math.sin(theta)
                cos_theta = math.cos(theta)
                sin_phi = math.sin(phi)
                cos_phi = math.cos(phi)
                s = theta/(2*math.pi) + 0.5
                tangent = 2*math.pi*cos_phi*glm.vec3(-sin_theta, cos_theta, 0)
                bitangent = math.pi*glm.vec3(-sin_phi*cos_theta, -sin_phi*sin_theta, cos_phi)
            t = int((position.z+1)/2)

            return glm.vec3(s, t, 0), tangent, bitangent

        theta = math.atan2(position.y, position.x)
        phi = math.atan(position.z/dem)
        sin_theta = math.sin(theta)
        cos_theta = math.cos(theta)
        sin_phi = math.sin(phi)
        cos_phi = math.cos(phi)

        s = theta/(2*math.pi) + 0.5
        t = phi/math.pi + 0.5

        tex_coord = glm.vec3(s, t, 0)
        tangent = 2*math.pi*cos_phi*glm.vec3(-sin_theta, cos_theta, 0)
        bitangent = math.pi*glm.vec3(-sin_phi*cos_theta, -sin_phi*sin_theta, cos_phi)
        return tex_coord, tangent, bitangent
    
    @staticmethod
    def __create_vertex(position, normal=None, radius=1, **kwargs):
        tex_coord, tangent, bitangent = Icosphere.__get_info(position, normal)
        return Vertex(
            position=radius*position,
            tangent=radius*tangent,
            bitangent=radius*bitangent,
            normal=position,
            tex_coord=tex_coord,
            **kwargs
        )

    @property
    def n_levels(self):
        return self.__n_levels
    
    @n_levels.setter
    @Mesh.param_setter
    def n_levels(self, n_levels:int):
        self.__n_levels = n_levels

    @property
    def radius(self):
        return self.__radius
    
    @radius.setter
    @Mesh.param_setter
    def radius(self, radius:float):
        self.__radius = radius



from ..Mesh import Mesh

from glass.utils import checktype
from glass import Vertex

import glm
import math

class Pyramid(Mesh):

    @checktype
    def __init__(self, n_sides:int=4, start_side:int=0, total_sides:int=None,
                 radius:float=1, height:float=1,
                 color:(glm.vec3,glm.vec4)=glm.vec4(0.396, 0.74151, 0.69102, 1), back_color:(glm.vec3,glm.vec4)=None,
                 normalize_tex_coord:bool=False, name:str="", block:bool=True):
        Mesh.__init__(self, name=name, block=block)
        self.__radius = radius
        self.__height = height
        self.__start_side = start_side
        self.__total_sides = total_sides
        self.__n_sides = n_sides
        self.__normalize_tex_coord = normalize_tex_coord
        self.start_building()

    def build(self):
        self.is_closed = True
        vertices = self.vertices
        indices = self.indices
        radius = self.__radius
        height = self.__height
        start_side = self.__start_side
        total_sides = self.__total_sides
        if total_sides is None:
            total_sides = self.__n_sides
        total_sides = min(self.__n_sides, total_sides)
        n_sides = self.__n_sides
        normalize_tex_coord = self.__normalize_tex_coord

        i_vertex = 0
        i_index = 0

        # 棱锥底面中心点
        vertex_bottom_center = Vertex()
        vertex_bottom_center.position = glm.vec3(0)
        vertex_bottom_center.normal = glm.vec3(0, 0, -1)
        vertex_bottom_center.tex_coord = glm.vec3(0.5, 0.5, 0)

        vertices[i_vertex] = vertex_bottom_center
        i_vertex += 1

        tex_coord_bottom_radius = 0.5 if normalize_tex_coord else radius
        
        internal_radius = radius * math.cos(math.pi/n_sides)
        side_height = math.sqrt(internal_radius**2 + height**2)
        half_side_width = radius * math.sin(math.pi/n_sides)

        s1 = 0.5 - half_side_width
        s2 = 0.5 + half_side_width
        t = side_height
        if normalize_tex_coord:
            t = 1
            s1 = 0.5 - half_side_width/side_height
            s2 = 0.5 + half_side_width/side_height

        for j in range(total_sides+1):
            theta = 2*math.pi*(j+start_side)/n_sides
            cos_theta = math.cos(theta)
            sin_theta = math.sin(theta)
            
            top = glm.vec3(0, 0, height)
            bottom = radius * glm.vec3(cos_theta, sin_theta, 0)

            vertex_top = Vertex()
            vertex_top.position = top
            vertex_top.tex_coord = glm.vec3(0.5, t, 0)

            vertex_side_bottom1 = Vertex()
            vertex_side_bottom1.position = bottom
            vertex_side_bottom1.tex_coord = glm.vec3(s1, 0, 0)

            vertex_side_bottom2 = Vertex()
            vertex_side_bottom2.position = bottom
            vertex_side_bottom2.tex_coord = glm.vec3(s2, 0, 0)

            vertex_bottom_bottom = Vertex()
            vertex_bottom_bottom.position = bottom
            vertex_bottom_bottom.normal = glm.vec3(0, 0, -1)
            vertex_bottom_bottom.tex_coord = glm.vec3(0.5+tex_coord_bottom_radius*cos_theta, 0.5+tex_coord_bottom_radius*sin_theta, 0)

            vertices[i_vertex] = vertex_top # 1 + 4*j
            i_vertex += 1

            vertices[i_vertex] = vertex_side_bottom1 # 1 + 4*j + 1
            i_vertex += 1

            vertices[i_vertex] = vertex_side_bottom2 # 1 + 4*j + 2
            i_vertex += 1

            vertices[i_vertex] = vertex_bottom_bottom # 1 + 4*j + 3
            i_vertex += 1

            if j > 0:
                # 侧面
                triangle = glm.uvec3(0, 0, 0)
                triangle[0] = 1 + 4*j + 1
                triangle[1] = 1 + 4*j
                triangle[2] = 1 + 4*j - 2
                indices[i_index] = triangle
                i_index += 1
                self.generate_temp_TBN(vertices[triangle[0]], vertices[triangle[1]], vertices[triangle[2]])

                # 底面
                triangle = glm.uvec3(0, 0, 0)
                triangle[0] = 1 + 4*j + 3
                triangle[1] = 1 + 4*j - 1
                triangle[2] = 0
                indices[i_index] = triangle
                i_index += 1
                self.generate_temp_TBN(vertices[triangle[0]], vertices[triangle[1]], vertices[triangle[2]])
            
                yield

        del vertices[i_vertex:]
        del indices[i_index:]

    @property
    def n_sides(self):
        return self.__n_sides
    
    @n_sides.setter
    @Mesh.param_setter
    def n_sides(self, n_sides):
        self.__n_sides = n_sides

    @property
    def start_side(self):
        return self.__start_side
    
    @start_side.setter
    @Mesh.param_setter
    def start_side(self, start_side:int):
        self.__start_side = start_side

    @property
    def total_sides(self):
        return self.__total_sides

    @total_sides.setter
    @Mesh.param_setter
    def total_sides(self, total_sides:int):
        self.__total_sides = total_sides

    @property
    def radius(self):
        return self.__radius
    
    @radius.setter
    @Mesh.param_setter
    def radius(self, radius:float):
        self.__radius = radius

    @property
    def height(self):
        return self.__height
    
    @height.setter
    @Mesh.param_setter
    def height(self, height:float):
        self.__height = height

    @property
    def normalize_tex_coord(self):
        return self.__normalize_tex_coord
    
    @normalize_tex_coord.setter
    @Mesh.param_setter
    def normalize_tex_coord(self, flag:bool):
        self.__normalize_tex_coord = flag



import glm
from OpenGL import GL
import numpy as np

from .GLInfo import GLInfo
from .EBO import EBO
from .SameTypeList import SameTypeList
from .utils import checktype
from .helper import sizeof
from .Increment import Increment

class Indices(SameTypeList):
	@checktype
	def __init__(self, _list:(list,np.ndarray)=None, draw_type:GLInfo.draw_types=GL.GL_STATIC_DRAW, dtype=None):
		SameTypeList.__init__(self, _list, glm.uvec3)
		
		self.stride = sizeof(glm.uvec3)
		self._ebo = EBO()
		self._draw_type = draw_type

		self._temp_buffer = None
		self._temp_buffer_changed = False
		self._temp_ebo = EBO()

	def _check_type(self, triangle):
		if not isinstance(triangle, glm.uvec3):
			raise TypeError("indices should be in type uvec3")

	@property
	def ebo(self):
		if self._temp_buffer is None:
			return self._ebo
		else:
			return self._temp_ebo

	@property
	def draw_type(self):
		return self._draw_type

	@draw_type.setter
	def draw_type(self, value):
		self._draw_type = value

	@property
	def temp_buffer(self):
		return self._temp_buffer
	
	@temp_buffer.setter
	def temp_buffer(self, temp_buffer:bytes):
		if self._temp_buffer == temp_buffer:
			return
		
		self._temp_buffer = temp_buffer
		self._temp_buffer_changed = True

	def _temp_apply(self):
		self._temp_ebo.bufferData(self._temp_buffer)
		self._temp_buffer_changed = False

	def _first_apply(self):
		if self._increment is not None or not self:
			return False

		self._increment = Increment(self)

		if self._ebo.nbytes == 0:
			self._ebo.bufferData(self.buffer, draw_type=self._draw_type)

		return True

	def _apply_increment(self):
		if not self._increment.is_changed:
			return
		
		patch = self._increment.patch()
		
		if patch["old_capacity"] == patch["new_capacity"]:
			for move in patch["move"]:
				self._ebo.memmove(move["old_start"]*self.stride, move["size"]*self.stride, move["new_start"]*self.stride)
		else:
			temp_ebo = EBO()
			temp_ebo.malloc(patch["new_capacity"]*self.stride, self._draw_type)
			self._ebo.copy_to(0, min(patch["old_size"], patch["new_size"])*self.stride, temp_ebo, 0)
			for move in patch["move"]:
				self._ebo.copy_to(move["old_start"]*self.stride, move["size"]*self.stride, temp_ebo, move["new_start"]*self.stride)

			self._ebo.delete()
			self._ebo = temp_ebo

		new_data = patch["new_data"]
		patch_update = patch["update"]
		len_patch_update = len(patch_update)

		if new_data and patch_update:
			temp_buffer = np.array(new_data).tobytes()
			if len_patch_update > 1:
				temp_ebo = EBO()
				temp_ebo.bufferData(temp_buffer, self._draw_type)

				for update in patch_update:
					dest_start = update["dest_start"]
					size = update["size"]
					src_start = update["src_start"]
					temp_ebo.copy_to(src_start*self.stride, size*self.stride, self._ebo, dest_start*self.stride)
			else:
				update = patch_update[0]
				dest_start = update["dest_start"]
				size = update["size"]
				self._ebo.bufferSubData(dest_start*self.stride, size*self.stride, temp_buffer)

	def _apply(self):
		if self._temp_buffer is not None:
			if self._temp_buffer_changed:
				self._temp_apply()
			return
		else:
			if self._temp_buffer_changed:
				self._temp_ebo.delete()
		
		self._check_in_items()
		
		success = False
		if self._increment is None:
			success = self._first_apply()
		else:
			self._apply_increment()
			success = True

		if success:
			self._ebo.bind()

	@property
	def buffer(self):
		return bytearray(self.ndarray.tobytes()) + bytearray((self.capacity - len(self))*self.stride)


from .PostProcessEffect import PostProcessEffect
from ..Frame import Frame

from glass import FBO, sampler2D, ShaderProgram, GLConfig
from glass.utils import checktype

from OpenGL import GL
import os

class BloomEffect(PostProcessEffect):

    @checktype
    def __init__(self):
        PostProcessEffect.__init__(self)
        
        self._src_width = 0
        self._src_height = 0

        self._bloom_down_fbo_list = []
        
        self._down_program = None
        self._up_program = None
        self._mix_fbo = None
        self._mix_program = None

        self.strength:float = 0.5
        self.threshold:float = 1
        self.blur_times:int = 6

    def need_pos_info(self)->bool:
        return False

    @property
    def down_program(self):
        if self._down_program is None:
            self._down_program = ShaderProgram()
            self._down_program.compile(Frame.draw_frame_vs)
            self._down_program.compile(os.path.dirname(os.path.abspath(__file__)) + "/../glsl/PostProcessEffects/bloom_downsampling.fs")
        
        return self._down_program
    
    @property
    def up_program(self):
        if self._up_program is None:
            self._up_program = ShaderProgram()
            self._up_program.compile(Frame.draw_frame_vs)
            self._up_program.compile(os.path.dirname(os.path.abspath(__file__)) + "/../glsl/PostProcessEffects/bloom_upsampling.fs")
        
        return self._up_program
    
    @property
    def mix_fbo(self):
        if self._mix_fbo is None:
            self._mix_fbo = FBO()
            self._mix_fbo.attach(0, sampler2D)
        
        return self._mix_fbo
    
    @property
    def mix_program(self):
        if self._mix_program is None:
            self._mix_program = ShaderProgram()
            self._mix_program.compile(Frame.draw_frame_vs)
            self._mix_program.compile(os.path.dirname(os.path.abspath(__file__)) + "/../glsl/PostProcessEffects/bloom_mix.fs")
        
        return self._mix_program
    
    def bloom_down_fbo(self, i:int)->FBO:
        if i < len(self._bloom_down_fbo_list):
            return self._bloom_down_fbo_list[i]
        else:
            return None

    def __get_bloom_image(self, screen_image:sampler2D):
        self.__update_fbo_list(screen_image.width, screen_image.height)
        
        with GLConfig.LocalConfig(cull_face=None, polygon_mode=GL.GL_FILL):
            current_index = 0
            for i in range(self.blur_times):
                down_fbo = self.bloom_down_fbo(i)
                if down_fbo is None:
                    break

                with down_fbo:
                    GLConfig.clear_buffers()
                    self.down_program["threshold"] = self.threshold
                    self.down_program["mip_level"] = i
                    self.down_program["screen_image"] = screen_image
                    self.down_program.draw_triangles(Frame.vertices, Frame.indices)
                screen_image = down_fbo.color_attachment(0)
                current_index = i

            with GLConfig.LocalConfig(blend=True,
                blend_src_rgb=GL.GL_ONE, blend_dest_rgb=GL.GL_ONE,
                blend_src_alpha=GL.GL_ONE, blend_dest_alpha=GL.GL_ONE):
                for i in range(current_index, -1, -1):
                    down_fbo = self.bloom_down_fbo(i)
                    with down_fbo:
                        self.up_program["filter_radius"] = 0.02 * self.strength
                        self.up_program["screen_image"] = screen_image
                        self.up_program.draw_triangles(Frame.vertices, Frame.indices)
                    screen_image = down_fbo.color_attachment(0)

        return screen_image

    def draw_to_active(self, screen_image:sampler2D)->None:
        bloom_image = self.__get_bloom_image(screen_image)
        with GLConfig.LocalConfig(cull_face=None, polygon_mode=GL.GL_FILL):
            self.mix_program["screen_image"] = screen_image
            self.mix_program["bloom_image"] = bloom_image
            self.mix_program.draw_triangles(Frame.vertices, Frame.indices)

    def apply(self, screen_image:sampler2D)->sampler2D:
        bloom_image = self.__get_bloom_image(screen_image)
        self.mix_fbo.resize(screen_image.width, screen_image.height)
        with GLConfig.LocalConfig(cull_face=None, polygon_mode=GL.GL_FILL):
            with self.mix_fbo:
                self.mix_program["screen_image"] = screen_image
                self.mix_program["bloom_image"] = bloom_image
                self.mix_program.draw_triangles(Frame.vertices, Frame.indices)
        return self.mix_fbo.color_attachment(0)

    def __update_fbo_list(self, src_width:int, src_height:int):
        if src_width == self._src_width and \
           src_height == self._src_height:
            return
        
        dest_width = 8
        dest_height = int(dest_width / src_width * src_height)

        width_list, height_list = [], []
        width, height = src_width, src_height

        is_width_down = (src_width > dest_width)
        scale = 0.5 if is_width_down else 2
        while (width > dest_width) == is_width_down:
            width = int(width * scale)
            width_list.append(width)

        is_height_down = (height > dest_height)
        scale = 0.5 if is_height_down else 2
        while (height > dest_height) == is_height_down:
            height = int(height * scale)
            height_list.append(height)

        len_width_list = len(width_list)
        len_height_list = len(height_list)
        delta_len = len_height_list - len_width_list

        if delta_len > 0:
            width_list.extend([width_list[-1]]*delta_len)
            len_width_list += delta_len
        elif delta_len < 0:
            height_list.extend([height_list[-1]]*(-delta_len))
            len_height_list -= delta_len

        len_bloom_down_fbo_list = len(self._bloom_down_fbo_list)
        while len_bloom_down_fbo_list < len_width_list:
            fbo = FBO()
            fbo.auto_clear = False
            fbo.attach(0, sampler2D)
            self._bloom_down_fbo_list.append(fbo)
            len_bloom_down_fbo_list += 1

        self.used_length = len_width_list

        for i in range(len_width_list):
            bloom_down_fbo = self._bloom_down_fbo_list[i]

            width = width_list[i]
            height = height_list[i]

            bloom_down_fbo.resize(width, height)

        self._src_width = src_width
        self._src_height = src_height



import numpy as np
import cv2
import re

def generate_standard_LUT(LUT_3D_size:int=64, dest_file_name:str=None):
    block_size = int(np.sqrt(LUT_3D_size))

    R = np.linspace(0, 1, LUT_3D_size)
    G = np.linspace(0, 1, LUT_3D_size)
    B = np.linspace(0, 1, LUT_3D_size)

    R, G = np.meshgrid(R, G)
    if block_size * block_size < LUT_3D_size:
        B = B.reshape((1, LUT_3D_size))
        R = np.tile(R, (1, LUT_3D_size))
        G = np.tile(G, (1, LUT_3D_size))
    else:
        B = B.reshape((block_size, block_size))
        R = np.tile(R, (block_size, block_size))
        G = np.tile(G, (block_size, block_size))

    B = np.repeat(B, LUT_3D_size, axis=0)
    B = np.repeat(B, LUT_3D_size, axis=1)
    
    result = np.stack((R, G, B), axis=-1)
    if dest_file_name is not None:
        out_image = (np.stack((B, G, R), axis=-1)*255).astype(np.uint8)
        cv2.imwrite(dest_file_name, out_image)
        
    return cv2.flip(result.astype(np.float32), 0)

def cube_info(cube_file_name:str):
    in_file = open(cube_file_name)
    content = in_file.read()
    in_file.close()

    lines = content.split("\n")
    data = []
    N = 0
    regx = re.compile(r"^\s*(?P<number1>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)\s+(?P<number2>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)\s+(?P<number3>[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)\s*$")
    for line in lines:
        line = line.strip()

        if line.startswith("#"):
            continue

        if line.startswith("LUT_3D_SIZE"):
            N = int(line[len("LUT_3D_SIZE"):].strip())
        else:
            match = re.fullmatch(regx, line)
            if match is not None:
                data.append([float(match["number1"]), float(match["number2"]), float(match["number3"])])

    data = np.array(data)
    return N, data

def apply_cube(cube:(str, tuple), src_image:(str, np.ndarray), dest_file_name:str=None):
    N, data = None, None
    if isinstance(cube, str):
        N, data = cube_info(cube)
    else:
        N, data = cube

    N2 = N * N
    
    if isinstance(src_image, str):
        if dest_file_name is None:
            dest_file_name = src_image

        src_image = cv2.flip(cv2.cvtColor(cv2.imread(src_image, cv2.IMREAD_UNCHANGED), cv2.COLOR_BGR2RGB), 0)

    if "int" in str(src_image.dtype):
        src_image = src_image / 255.0

    src_image *= (N - 1)
    R = src_image[:, :, 0]
    G = src_image[:, :, 1]
    B = src_image[:, :, 2]

    r = {}
    r[0] = np.floor(R).astype(np.uint32)
    r[1] = np.ceil(R).astype(np.uint32)
    R_rear = R - r[0]
    R_rear = np.stack((R_rear, R_rear, R_rear), axis=-1)

    g = {}
    g[0] = np.floor(G).astype(np.uint32)
    g[1] = np.ceil(G).astype(np.uint32)
    G_rear = G - g[0]
    G_rear = np.stack((G_rear, G_rear, G_rear), axis=-1)

    b = {}
    b[0] = np.floor(B).astype(np.uint32)
    b[1] = np.ceil(B).astype(np.uint32)
    B_rear = B - b[0]
    B_rear = np.stack((B_rear, B_rear, B_rear), axis=-1)

    values = {}
    for i in range(2):
        for j in range(2):
            for k in range(2):
                index = r[i] + N*g[j] + N2*b[k]
                values[i, j, k] = data[index, :]

    for j in range(2):
        for k in range(2):
            values[0.5, j, k] = (1-R_rear) * values[0, j, k] + R_rear * values[1, j, k]

    values[0.5, 0.5, 0] = (1 - G_rear) * values[0.5, 0, 0] + G_rear * values[0.5, 1, 0]
    values[0.5, 0.5, 1] = (1 - G_rear) * values[0.5, 0, 1] + G_rear * values[0.5, 1, 1]
    values[0.5, 0.5, 0.5] = (1 - B_rear) * values[0.5, 0.5, 0] + B_rear * values[0.5, 0.5, 1]
    
    result = values[0.5, 0.5, 0.5]
    if dest_file_name is not None:
        out_image = cv2.flip(cv2.cvtColor((np.clip(result, 0, 1)*255).astype(np.uint8), cv2.COLOR_RGB2BGR), 0)
        cv2.imwrite(dest_file_name, out_image)

    return result.astype(np.float32)

def cube_to_LUT(cube_file_name:str, dest_file_name:str=None):
    N, data = cube_info(cube_file_name)
    return apply_cube((N,data), generate_standard_LUT(N), dest_file_name)



from .PostProcessEffect import PostProcessEffect
from ..Frame import Frame

from glass import FBO, ShaderProgram, sampler2D, GLConfig, sampler2DArray, samplerCube, GLInfo
from glass.utils import checktype
from glass.helper import get_channels

from OpenGL import GL
import glm
import os

class GaussBlur(PostProcessEffect):

    __array_programs = {}
    __program = None
    __cube_program = None

    @staticmethod
    def program():
        if GaussBlur.__program is None:
            GaussBlur.__program = ShaderProgram()
            GaussBlur.__program.compile(Frame.draw_frame_vs)
            GaussBlur.__program.compile(os.path.dirname(os.path.abspath(__file__)) + "/../glsl/PostProcessEffects/gauss_blur.fs")
        return GaussBlur.__program
    
    @staticmethod
    def cube_program():
        if GaussBlur.__cube_program is None:
            GaussBlur.__cube_program = ShaderProgram()
            GaussBlur.__cube_program.compile(Frame.draw_frame_vs)
            GaussBlur.__cube_program.compile(Frame.draw_frame_array_gs(6))
            GaussBlur.__cube_program.compile(os.path.dirname(os.path.abspath(__file__)) + "/../glsl/PostProcessEffects/gauss_cube_blur.fs")
        return GaussBlur.__cube_program
    
    @checktype
    def __init__(self, kernel_shape:(int,tuple)=32, sigma:(float,tuple)=0, internal_format:GLInfo.internal_formats=GL.GL_RGBA32F):
        PostProcessEffect.__init__(self)
        
        self.__kernel_shape = glm.uvec2()
        self.__sigma = glm.vec2()
        self.__channels = get_channels(internal_format)
        self.__internal_format = internal_format

        if isinstance(kernel_shape, tuple):
            self.__kernel_shape.x, self.__kernel_shape.y = kernel_shape
        else:
            self.__kernel_shape.x, self.__kernel_shape.y = kernel_shape, kernel_shape

        if isinstance(sigma, tuple):
            self.__sigma.x, self.__sigma.y = sigma
        else:
            self.__sigma.x, self.__sigma.y = sigma, sigma

        self._horizontal_fbo = None
        self._vertical_fbo = None
        self._horizontal_array_fbo = None
        self._vertical_array_fbo = None
        self._horizontal_cube_fbo = None
        self._vertical_cube_fbo = None

    def need_pos_info(self):
        return False

    @property
    def horizontal_fbo(self):
        if self._horizontal_fbo is None:
            self._horizontal_fbo = FBO()
            self._horizontal_fbo.attach(0, sampler2D, self.__internal_format)
        return self._horizontal_fbo

    @property
    def vertical_fbo(self):
        if self._vertical_fbo is None:
            self._vertical_fbo = FBO()
            self._vertical_fbo.attach(0, sampler2D, self.__internal_format)
        return self._vertical_fbo

    @property
    def horizontal_array_fbo(self):
        if self._horizontal_array_fbo is None:
            self._horizontal_array_fbo = FBO()
            self._horizontal_array_fbo.attach(0, sampler2DArray, self.__internal_format)
        return self._horizontal_array_fbo

    @property
    def vertical_array_fbo(self):
        if self._vertical_array_fbo is None:
            self._vertical_array_fbo = FBO()
            self._vertical_array_fbo.attach(0, sampler2DArray, self.__internal_format)
        return self._vertical_array_fbo

    @property
    def horizontal_cube_fbo(self):
        if self._horizontal_cube_fbo is None:
            self._horizontal_cube_fbo = FBO()
            self._horizontal_cube_fbo.attach(0, samplerCube, self.__internal_format)
        return self._horizontal_cube_fbo

    @property
    def vertical_cube_fbo(self):
        if self._vertical_cube_fbo is None:
            self._vertical_cube_fbo = FBO()
            self._vertical_cube_fbo.attach(0, samplerCube, self.__internal_format)
        return self._vertical_cube_fbo

    def apply(self, screen_image:(sampler2D,sampler2DArray,samplerCube))->(sampler2D,sampler2DArray,samplerCube):
        if self.__sigma.x == 0:
            self.__sigma.x = 0.3 * ((self.__kernel_shape.x-1)*0.5 - 1) + 0.8
        if self.__sigma.y == 0:
            self.__sigma.y = 0.3 * ((self.__kernel_shape.y-1)*0.5 - 1) + 0.8
        
        if isinstance(screen_image, sampler2D):
            self.horizontal_fbo.resize(screen_image.width, screen_image.height)
            self.vertical_fbo.resize(screen_image.width, screen_image.height)

            program = GaussBlur.program()
            with GLConfig.LocalConfig(cull_face=None, polygon_mode=GL.GL_FILL):

                program["kernel_shape"] = self.__kernel_shape
                program["sigma"] = self.__sigma
                program["channels"] = self.__channels

                with self.horizontal_fbo:
                    program["screen_image"] = screen_image
                    program["horizontal"] = True
                    program.draw_triangles(Frame.vertices, Frame.indices)

                with self.vertical_fbo:
                    program["screen_image"] = self.horizontal_fbo.color_attachment(0)
                    program["horizontal"] = False
                    program.draw_triangles(Frame.vertices, Frame.indices)

            return self.vertical_fbo.color_attachment(0)
        elif isinstance(screen_image, sampler2DArray):
            self.horizontal_array_fbo.resize(screen_image.width, screen_image.height, layers=screen_image.layers)
            self.vertical_array_fbo.resize(screen_image.width, screen_image.height, layers=screen_image.layers)

            program = GaussBlur.array_program(screen_image.layers)

            with GLConfig.LocalConfig(cull_face=None, polygon_mode=GL.GL_FILL):
                
                program["kernel_shape"] = self.__kernel_shape
                program["sigma"] = self.__sigma
                program["channels"] = self.__channels

                with self.horizontal_array_fbo:
                    program["screen_image"] = screen_image
                    program["horizontal"] = True
                    program.draw_triangles(Frame.vertices, Frame.indices)

                with self.vertical_array_fbo:
                    program["screen_image"] = self.horizontal_array_fbo.color_attachment(0)
                    program["horizontal"] = False
                    program.draw_triangles(Frame.vertices, Frame.indices)

            return self.vertical_array_fbo.color_attachment(0)
        elif isinstance(screen_image, samplerCube):
            self.horizontal_cube_fbo.resize(screen_image.width, screen_image.height)
            self.vertical_cube_fbo.resize(screen_image.width, screen_image.height)

            program = GaussBlur.cube_program()

            with GLConfig.LocalConfig(cull_face=None, polygon_mode=GL.GL_FILL):
                
                program["kernel_shape"] = self.__kernel_shape
                program["sigma"] = self.__sigma
                program["channels"] = self.__channels

                with self.horizontal_cube_fbo:
                    program["screen_image"] = screen_image
                    program["horizontal"] = True
                    program.draw_triangles(Frame.vertices, Frame.indices)

                with self.vertical_cube_fbo:
                    program["screen_image"] = self.horizontal_cube_fbo.color_attachment(0)
                    program["horizontal"] = False
                    program.draw_triangles(Frame.vertices, Frame.indices)

            return self.vertical_cube_fbo.color_attachment(0)
    
    def draw_to_active(self, screen_image:sampler2D) -> None:
        self.horizontal_fbo.resize(screen_image.width, screen_image.height)
        self.vertical_fbo.resize(screen_image.width, screen_image.height)

        if self.__sigma.x == 0:
            self.__sigma.x = ((self.__kernel_shape.x-1)*0.5 - 1)/3
        if self.__sigma.y == 0:
            self.__sigma.y = ((self.__kernel_shape.y-1)*0.5 - 1)/3

        program = GaussBlur.program()

        with GLConfig.LocalConfig(cull_face=None, polygon_mode=GL.GL_FILL):

            program["kernel_shape"] = self.__kernel_shape
            program["sigma"] = self.__sigma
            program["channels"] = self.__channels

            with self.horizontal_fbo:
                program["screen_image"] = screen_image
                program["horizontal"] = True
                program.draw_triangles(Frame.vertices, Frame.indices)

            GLConfig.clear_buffers()
            program["screen_image"] = self.horizontal_fbo.color_attachment(0)
            program["horizontal"] = False
            program.draw_triangles(Frame.vertices, Frame.indices)

    @property
    def kernel_shape(self)->tuple:
        return (self.__kernel_shape.x, self.__kernel_shape.y)
    
    @kernel_shape.setter
    @checktype
    def kernel_shape(self, shape:(int, tuple)):
        if isinstance(shape, tuple):
            self.__kernel_shape.x, self.__kernel_shape.y = shape
        else:
            self.__kernel_shape.x, self.__kernel_shape.y = shape, shape
    
    @property
    def kernel_width(self)->int:
        return self.__kernel_shape.x
    
    @kernel_width.setter
    @checktype
    def kernel_width(self, width:int):
        self.__kernel_shape.x = width
    
    @property
    def kernel_height(self)->int:
        return self.__kernel_shape.y
    
    @kernel_height.setter
    @checktype
    def kernel_height(self, height:int):
        self.__kernel_shape.y = height

    @property
    def sigma(self)->tuple:
        return (self.__sigma.x, self.__sigma.y)
    
    @sigma.setter
    @checktype
    def sigma(self, sigma:(float,tuple)):
        if isinstance(sigma, tuple):
            self.__sigma.x, self.__sigma.y = sigma
        else:
            self.__sigma.x, self.__sigma.y = sigma, sigma

    @property
    def sigma_x(self)->float:
        return self.__sigma.x
    
    @sigma_x.setter
    def sigma_x(self, value:float):
        self.__sigma.x = value

    @property
    def sigma_y(self)->float:
        return self.__sigma.y
    
    @sigma_y.setter
    def sigma_y(self, value:float):
        self.__sigma.y = value
    
    @staticmethod
    def array_program(layers):
        if layers in GaussBlur.__array_programs:
            return GaussBlur.__array_programs[layers]

        program = ShaderProgram()
        program.compile(Frame.draw_frame_vs)
        program.compile(Frame.draw_frame_array_gs(layers))
        program.compile(os.path.dirname(os.path.abspath(__file__)) + "/../glsl/PostProcessEffects/gauss_array_blur.fs")

        GaussBlur.__array_programs[layers] = program

        return program


from glass.utils import di
from enum import Flag, IntEnum

import glm

class Manipulator:

    class MouseButton(Flag):
        LeftButton = 1
        RightButton = 2      
        MiddleButton = 4     
        XButton1 = 8
        XButton2 = 16        
        TaskButton = 32      
        ExtraButton4 = 64    
        ExtraButton5 = 128   
        ExtraButton6 = 256   
        ExtraButton7 = 512   
        ExtraButton8 = 1024  
        ExtraButton9 = 2048  
        ExtraButton10 = 4096 
        ExtraButton11 = 8192 
        ExtraButton12 = 16384
        ExtraButton13 = 32768
        ExtraButton14 = 65536
        ExtraButton15 = 131072
        ExtraButton16 = 262144
        ExtraButton17 = 524288
        ExtraButton18 = 1048576
        ExtraButton19 = 2097152
        ExtraButton20 = 4194304
        ExtraButton21 = 8388608
        ExtraButton22 = 16777216
        ExtraButton23 = 33554432
        ExtraButton24 = 67108864

    class Key(IntEnum):
        Key_Escape = 16777216   
        Key_Tab = 16777217      
        Key_Backtab = 16777218  
        Key_Backspace = 16777219
        Key_Return = 16777220   
        Key_Enter = 16777221    
        Key_Insert = 16777222   
        Key_Delete = 16777223   
        Key_Pause = 16777224    
        Key_Print = 16777225    
        Key_SysReq = 16777226   
        Key_Clear = 16777227    
        Key_Home = 16777232     
        Key_End = 16777233      
        Key_Left = 16777234
        Key_Up = 16777235
        Key_Right = 16777236
        Key_Down = 16777237
        Key_PageUp = 16777238
        Key_PageDown = 16777239
        Key_Shift = 16777248
        Key_Control = 16777249
        Key_Meta = 16777250
        Key_Alt = 16777251
        Key_CapsLock = 16777252
        Key_NumLock = 16777253
        Key_ScrollLock = 16777254
        Key_F1 = 16777264
        Key_F2 = 16777265
        Key_F3 = 16777266
        Key_F4 = 16777267
        Key_F5 = 16777268
        Key_F6 = 16777269
        Key_F7 = 16777270
        Key_F8 = 16777271
        Key_F9 = 16777272
        Key_F10 = 16777273
        Key_F11 = 16777274
        Key_F12 = 16777275
        Key_F13 = 16777276
        Key_F14 = 16777277
        Key_F15 = 16777278
        Key_F16 = 16777279
        Key_F17 = 16777280
        Key_F18 = 16777281
        Key_F19 = 16777282
        Key_F20 = 16777283
        Key_F21 = 16777284
        Key_F22 = 16777285
        Key_F23 = 16777286
        Key_F24 = 16777287
        Key_F25 = 16777288
        Key_F26 = 16777289
        Key_F27 = 16777290
        Key_F28 = 16777291
        Key_F29 = 16777292
        Key_F30 = 16777293
        Key_F31 = 16777294
        Key_F32 = 16777295
        Key_F33 = 16777296
        Key_F34 = 16777297
        Key_F35 = 16777298
        Key_Super_L = 16777299
        Key_Super_R = 16777300
        Key_Menu = 16777301
        Key_Hyper_L = 16777302
        Key_Hyper_R = 16777303
        Key_Help = 16777304
        Key_Direction_L = 16777305
        Key_Direction_R = 16777312
        Key_Space = 32
        Key_Exclam = 33
        Key_QuoteDbl = 34
        Key_NumberSign = 35
        Key_Dollar = 36
        Key_Percent = 37
        Key_Ampersand = 38
        Key_Apostrophe = 39
        Key_ParenLeft = 40
        Key_ParenRight = 41
        Key_Asterisk = 42
        Key_Plus = 43
        Key_Comma = 44
        Key_Minus = 45
        Key_Period = 46
        Key_Slash = 47
        Key_0 = 48
        Key_1 = 49
        Key_2 = 50
        Key_3 = 51
        Key_4 = 52
        Key_5 = 53
        Key_6 = 54
        Key_7 = 55
        Key_8 = 56
        Key_9 = 57
        Key_Colon = 58
        Key_Semicolon = 59
        Key_Less = 60
        Key_Equal = 61
        Key_Greater = 62
        Key_Question = 63
        Key_At = 64
        Key_A = 65
        Key_B = 66
        Key_C = 67
        Key_D = 68
        Key_E = 69
        Key_F = 70
        Key_G = 71
        Key_H = 72
        Key_I = 73
        Key_J = 74
        Key_K = 75
        Key_L = 76
        Key_M = 77
        Key_N = 78
        Key_O = 79
        Key_P = 80
        Key_Q = 81
        Key_R = 82
        Key_S = 83
        Key_T = 84
        Key_U = 85
        Key_V = 86
        Key_W = 87
        Key_X = 88
        Key_Y = 89
        Key_Z = 90
        Key_BracketLeft = 91
        Key_Backslash = 92
        Key_BracketRight = 93
        Key_AsciiCircum = 94
        Key_Underscore = 95
        Key_QuoteLeft = 96
        Key_BraceLeft = 123
        Key_Bar = 124
        Key_BraceRight = 125
        Key_AsciiTilde = 126
        Key_nobreakspace = 160
        Key_exclamdown = 161
        Key_cent = 162
        Key_sterling = 163
        Key_currency = 164
        Key_yen = 165
        Key_brokenbar = 166
        Key_section = 167
        Key_diaeresis = 168
        Key_copyright = 169
        Key_ordfeminine = 170
        Key_guillemotleft = 171
        Key_notsign = 172
        Key_hyphen = 173
        Key_registered = 174
        Key_macron = 175
        Key_degree = 176
        Key_plusminus = 177
        Key_twosuperior = 178
        Key_threesuperior = 179
        Key_acute = 180
        Key_mu = 181
        Key_paragraph = 182
        Key_periodcentered = 183
        Key_cedilla = 184
        Key_onesuperior = 185
        Key_masculine = 186
        Key_guillemotright = 187
        Key_onequarter = 188
        Key_onehalf = 189
        Key_threequarters = 190
        Key_questiondown = 191
        Key_Agrave = 192
        Key_Aacute = 193
        Key_Acircumflex = 194
        Key_Atilde = 195
        Key_Adiaeresis = 196
        Key_Aring = 197
        Key_AE = 198
        Key_Ccedilla = 199
        Key_Egrave = 200
        Key_Eacute = 201
        Key_Ecircumflex = 202
        Key_Ediaeresis = 203
        Key_Igrave = 204
        Key_Iacute = 205
        Key_Icircumflex = 206
        Key_Idiaeresis = 207
        Key_ETH = 208
        Key_Ntilde = 209
        Key_Ograve = 210
        Key_Oacute = 211
        Key_Ocircumflex = 212
        Key_Otilde = 213
        Key_Odiaeresis = 214
        Key_multiply = 215
        Key_Ooblique = 216
        Key_Ugrave = 217
        Key_Uacute = 218
        Key_Ucircumflex = 219
        Key_Udiaeresis = 220
        Key_Yacute = 221
        Key_THORN = 222
        Key_ssharp = 223
        Key_division = 247
        Key_ydiaeresis = 255
        Key_AltGr = 16781571
        Key_Multi_key = 16781600
        Key_Codeinput = 16781623
        Key_SingleCandidate = 16781628
        Key_MultipleCandidate = 16781629
        Key_PreviousCandidate = 16781630
        Key_Mode_switch = 16781694
        Key_Kanji = 16781601
        Key_Muhenkan = 16781602
        Key_Henkan = 16781603
        Key_Romaji = 16781604
        Key_Hiragana = 16781605
        Key_Katakana = 16781606
        Key_Hiragana_Katakana = 16781607
        Key_Zenkaku = 16781608
        Key_Hankaku = 16781609
        Key_Zenkaku_Hankaku = 16781610
        Key_Touroku = 16781611
        Key_Massyo = 16781612
        Key_Kana_Lock = 16781613
        Key_Kana_Shift = 16781614
        Key_Eisu_Shift = 16781615
        Key_Eisu_toggle = 16781616
        Key_Hangul = 16781617
        Key_Hangul_Start = 16781618
        Key_Hangul_End = 16781619
        Key_Hangul_Hanja = 16781620
        Key_Hangul_Jamo = 16781621
        Key_Hangul_Romaja = 16781622
        Key_Hangul_Jeonja = 16781624
        Key_Hangul_Banja = 16781625
        Key_Hangul_PreHanja = 16781626
        Key_Hangul_PostHanja = 16781627
        Key_Hangul_Special = 16781631
        Key_Dead_Grave = 16781904
        Key_Dead_Acute = 16781905
        Key_Dead_Circumflex = 16781906
        Key_Dead_Tilde = 16781907
        Key_Dead_Macron = 16781908
        Key_Dead_Breve = 16781909
        Key_Dead_Abovedot = 16781910
        Key_Dead_Diaeresis = 16781911
        Key_Dead_Abovering = 16781912
        Key_Dead_Doubleacute = 16781913
        Key_Dead_Caron = 16781914
        Key_Dead_Cedilla = 16781915
        Key_Dead_Ogonek = 16781916
        Key_Dead_Iota = 16781917
        Key_Dead_Voiced_Sound = 16781918
        Key_Dead_Semivoiced_Sound = 16781919
        Key_Dead_Belowdot = 16781920
        Key_Dead_Hook = 16781921
        Key_Dead_Horn = 16781922
        Key_Back = 16777313
        Key_Forward = 16777314
        Key_Stop = 16777315
        Key_Refresh = 16777316
        Key_VolumeDown = 16777328
        Key_VolumeMute = 16777329
        Key_VolumeUp = 16777330
        Key_BassBoost = 16777331
        Key_BassUp = 16777332
        Key_BassDown = 16777333
        Key_TrebleUp = 16777334
        Key_TrebleDown = 16777335
        Key_MediaPlay = 16777344
        Key_MediaStop = 16777345
        Key_MediaPrevious = 16777346
        Key_MediaNext = 16777347
        Key_MediaRecord = 16777348
        Key_HomePage = 16777360
        Key_Favorites = 16777361
        Key_Search = 16777362
        Key_Standby = 16777363
        Key_OpenUrl = 16777364
        Key_LaunchMail = 16777376
        Key_LaunchMedia = 16777377
        Key_Launch0 = 16777378
        Key_Launch1 = 16777379
        Key_Launch2 = 16777380
        Key_Launch3 = 16777381
        Key_Launch4 = 16777382
        Key_Launch5 = 16777383
        Key_Launch6 = 16777384
        Key_Launch7 = 16777385
        Key_Launch8 = 16777386
        Key_Launch9 = 16777387
        Key_LaunchA = 16777388
        Key_LaunchB = 16777389
        Key_LaunchC = 16777390
        Key_LaunchD = 16777391
        Key_LaunchE = 16777392
        Key_LaunchF = 16777393
        Key_MediaLast = 16842751
        Key_Select = 16842752
        Key_Yes = 16842753
        Key_No = 16842754
        Key_Context1 = 17825792
        Key_Context2 = 17825793
        Key_Context3 = 17825794
        Key_Context4 = 17825795
        Key_Call = 17825796
        Key_Hangup = 17825797
        Key_Flip = 17825798
        Key_unknown = 33554431
        Key_Execute = 16908291
        Key_Printer = 16908290
        Key_Play = 16908293
        Key_Sleep = 16908292
        Key_Zoom = 16908294
        Key_Cancel = 16908289
        Key_MonBrightnessUp = 16777394
        Key_MonBrightnessDown = 16777395
        Key_KeyboardLightOnOff = 16777396
        Key_KeyboardBrightnessUp = 16777397
        Key_KeyboardBrightnessDown = 16777398
        Key_PowerOff = 16777399
        Key_WakeUp = 16777400
        Key_Eject = 16777401
        Key_ScreenSaver = 16777402
        Key_WWW = 16777403
        Key_Memo = 16777404
        Key_LightBulb = 16777405
        Key_Shop = 16777406
        Key_History = 16777407
        Key_AddFavorite = 16777408
        Key_HotLinks = 16777409
        Key_BrightnessAdjust = 16777410
        Key_Finance = 16777411
        Key_Community = 16777412
        Key_AudioRewind = 16777413
        Key_BackForward = 16777414
        Key_ApplicationLeft = 16777415
        Key_ApplicationRight = 16777416
        Key_Book = 16777417
        Key_CD = 16777418
        Key_Calculator = 16777419
        Key_ToDoList = 16777420
        Key_ClearGrab = 16777421
        Key_Close = 16777422
        Key_Copy = 16777423
        Key_Cut = 16777424
        Key_Display = 16777425
        Key_DOS = 16777426
        Key_Documents = 16777427
        Key_Excel = 16777428
        Key_Explorer = 16777429
        Key_Game = 16777430
        Key_Go = 16777431
        Key_iTouch = 16777432
        Key_LogOff = 16777433
        Key_Market = 16777434
        Key_Meeting = 16777435
        Key_MenuKB = 16777436
        Key_MenuPB = 16777437
        Key_MySites = 16777438
        Key_News = 16777439
        Key_OfficeHome = 16777440
        Key_Option = 16777441
        Key_Paste = 16777442
        Key_Phone = 16777443
        Key_Calendar = 16777444
        Key_Reply = 16777445
        Key_Reload = 16777446
        Key_RotateWindows = 16777447
        Key_RotationPB = 16777448
        Key_RotationKB = 16777449
        Key_Save = 16777450
        Key_Send = 16777451
        Key_Spell = 16777452
        Key_SplitScreen = 16777453
        Key_Support = 16777454
        Key_TaskPane = 16777455
        Key_Terminal = 16777456
        Key_Tools = 16777457
        Key_Travel = 16777458
        Key_Video = 16777459
        Key_Word = 16777460
        Key_Xfer = 16777461
        Key_ZoomIn = 16777462
        Key_ZoomOut = 16777463
        Key_Away = 16777464
        Key_Messenger = 16777465
        Key_WebCam = 16777466
        Key_MailForward = 16777467
        Key_Pictures = 16777468
        Key_Music = 16777469
        Key_Battery = 16777470
        Key_Bluetooth = 16777471
        Key_WLAN = 16777472
        Key_UWB = 16777473
        Key_AudioForward = 16777474
        Key_AudioRepeat = 16777475
        Key_AudioRandomPlay = 16777476
        Key_Subtitle = 16777477
        Key_AudioCycleTrack = 16777478
        Key_Time = 16777479
        Key_Hibernate = 16777480
        Key_View = 16777481
        Key_TopMenu = 16777482
        Key_PowerDown = 16777483
        Key_Suspend = 16777484
        Key_ContrastAdjust = 16777485
        Key_MediaPause = 16777349
        Key_MediaTogglePlayPause = 16777350
        Key_LaunchG = 16777486
        Key_LaunchH = 16777487
        Key_ToggleCallHangup = 17825799
        Key_VoiceDial = 17825800
        Key_LastNumberRedial = 17825801
        Key_Camera = 17825824
        Key_CameraFocus = 17825825
        Key_TouchpadToggle = 16777488
        Key_TouchpadOn = 16777489
        Key_TouchpadOff = 16777490
        Key_MicMute = 16777491
        Key_Red = 16777492
        Key_Green = 16777493
        Key_Yellow = 16777494
        Key_Blue = 16777495
        Key_ChannelUp = 16777496
        Key_ChannelDown = 16777497
        Key_Guide = 16777498
        Key_Info = 16777499
        Key_Settings = 16777500
        Key_Exit = 16908298
        Key_MicVolumeUp = 16777501
        Key_MicVolumeDown = 16777502
        Key_New = 16777504
        Key_Open = 16777505
        Key_Find = 16777506
        Key_Undo = 16777507
        Key_Redo = 16777508
        Key_Dead_Stroke = 16781923
        Key_Dead_Abovecomma = 16781924
        Key_Dead_Abovereversedcomma = 16781925
        Key_Dead_Doublegrave = 16781926
        Key_Dead_Belowring = 16781927
        Key_Dead_Belowmacron = 16781928
        Key_Dead_Belowcircumflex = 16781929
        Key_Dead_Belowtilde = 16781930
        Key_Dead_Belowbreve = 16781931
        Key_Dead_Belowdiaeresis = 16781932
        Key_Dead_Invertedbreve = 16781933
        Key_Dead_Belowcomma = 16781934
        Key_Dead_Currency = 16781935
        Key_Dead_a = 16781952
        Key_Dead_A = 16781953
        Key_Dead_e = 16781954
        Key_Dead_E = 16781955
        Key_Dead_i = 16781956
        Key_Dead_I = 16781957
        Key_Dead_o = 16781958
        Key_Dead_O = 16781959
        Key_Dead_u = 16781960
        Key_Dead_U = 16781961
        Key_Dead_Small_Schwa = 16781962
        Key_Dead_Capital_Schwa = 16781963
        Key_Dead_Greek = 16781964
        Key_Dead_Lowline = 16781968
        Key_Dead_Aboveverticalline = 16781969
        Key_Dead_Belowverticalline = 16781970
        Key_Dead_Longsolidusoverlay = 16781971

    def __init__(self):
        self._camera_id = id(None)

    @property
    def camera(self):        
        return di(self._camera_id)

    def startup(self)->bool:
        return False

    def on_mouse_pressed(self, button:MouseButton, screen_pos:glm.vec2, global_pos:glm.vec2)->bool:
        return False

    def on_mouse_released(self, button:MouseButton, screen_pos:glm.vec2, global_pos:glm.vec2)->bool:
        return False
    
    def on_mouse_double_clicked(self, button:MouseButton, screen_pos:glm.vec2, global_pos:glm.vec2)->bool:
        return False

    def on_mouse_moved(self, screen_pos:glm.vec2, global_pos:glm.vec2)->bool:
        return False

    def on_wheel_scrolled(self, angle:glm.vec2, screen_pos:glm.vec2, global_pos:glm.vec2)->bool:
        return False

    def on_key_pressed(self, key:Key)->bool:
        return False

    def on_key_released(self, key:Key)->bool:
        return False

    def on_key_repeated(self, keys:set)->bool:
        return False


import glm
import math
from enum import Enum
from OpenGL import GL

from glass.utils import checktype
from glass import GLInfo

class Fog:

    class Mode(Enum):
        Linear = GL.GL_LINEAR
        Exp = GL.GL_EXP
        Exp2 = GL.GL_EXP2

    def __init__(self):
        self._mode = Fog.Mode.Exp
        self._color = glm.vec3(1, 1, 1)
        self._extinction_density = 0
        self._inscattering_density = 0

    def apply(self, color:(glm.vec3, glm.vec4), camera_pos:glm.vec3, frag_pos:glm.vec3):
        if self.extinction_density < 1E-6 and \
           self.inscattering_density < 1E-6:
            return color

        d = glm.length(frag_pos - camera_pos)
        extinction = 1
        inscattering = 1
        if self.mode == Fog.Mode.Linear:
            extinction = max(0, 1 - self.extinction_density*d)
            inscattering = max(0, 1 - self.inscattering_density*d)
        elif self.mode == Fog.Mode.Exp:
            extinction = math.exp(-self.extinction_density*d)
            inscattering = math.exp(-self.inscattering_density*d)
        elif self.mode == Fog.Mode.Exp2:
            extinction = math.exp(-self.extinction_density*d*d)
            inscattering = math.exp(-self.inscattering_density*d*d)

        self_color = self.color
        if isinstance(color, glm.vec4):
            self_color = glm.vec4(self_color, 1)

        return color * extinction + self_color * (1 - inscattering)

    @property
    def mode(self)->Mode:
        return self._mode
    
    @mode.setter
    @checktype
    def mode(self, mode:[Mode.Linear, Mode.Exp, Mode.Exp2, *GLInfo.fog_modes]):
        if mode in GLInfo.fog_modes:
            mode = Fog.Mode(mode)

        self._mode = mode

    @property
    def color(self)->glm.vec3:
        return self._color
    
    @color.setter
    @checktype
    def color(self, color:glm.vec3):
        self._color = color

    @property
    def density(self)->float:
        return self._extinction_density
    
    @density.setter
    @checktype
    def density(self, density:float):
        self._extinction_density = density
        self._inscattering_density = density

    @property
    def extinction_density(self)->float:
        return self._extinction_density
    
    @extinction_density.setter
    @checktype
    def extinction_density(self, density:float):
        self._extinction_density = density

    @property
    def inscattering_density(self)->float:
        return self._inscattering_density
    
    @inscattering_density.setter
    @checktype
    def inscattering_density(self, density:float):
        self._inscattering_density = density
        


from ..Mesh import Mesh

from glass.utils import checktype
from glass import Vertex

import glm
import math

class CylinderSide(Mesh):

    @checktype
    def __init__(self, radius:float=1, height:float=1,
                 start_angle:float=0, span_angle:float=360, n_divide:int=100,
                 color:(glm.vec3,glm.vec4)=glm.vec4(0.396, 0.74151, 0.69102, 1), back_color:(glm.vec3,glm.vec4)=None,
                 normalize_tex_coord=False,
                 name:str="", block=True):
        Mesh.__init__(self, color=color, back_color=back_color, name=name, block=block)
        self.__radius = radius
        self.__height = height
        self.__start_angle = start_angle
        self.__span_angle = span_angle
        self.__n_divide = n_divide
        self.__normalize_tex_coord = normalize_tex_coord
        self.start_building()

    def build(self):
        self.is_closed = False
        self.self_calculated_normal = True

        vertices = self.vertices
        indices = self.indices

        i_vertex = 0
        i_index = 0

        radius = self.__radius
        height = self.__height
        start_angle = self.__start_angle/180*math.pi
        span_angle = self.__span_angle/180*math.pi
        n_divide = self.__n_divide
        normalize_tex_coord = self.__normalize_tex_coord

        t = 1 if normalize_tex_coord else height

        for j in range(n_divide):
            theta =  start_angle + span_angle * j / (n_divide - 1)
            cos_theta = math.cos(theta)
            sin_theta = math.sin(theta)

            s = radius*theta/height*t

            top = glm.vec3(radius*cos_theta, radius*sin_theta, height)
            bottom = glm.vec3(radius*cos_theta, radius*sin_theta, 0)
            normal = glm.vec3(cos_theta, sin_theta, 0)

            vertex_top = Vertex()
            vertex_top.position = top
            vertex_top.normal = normal
            vertex_top.tex_coord = glm.vec3(s, t, 0)

            vertex_bottom = Vertex()
            vertex_bottom.position = bottom
            vertex_bottom.normal = normal
            vertex_bottom.tex_coord = glm.vec3(s, 0, 0)

            vertices[i_vertex] = vertex_top # 2*j
            i_vertex += 1

            vertices[i_vertex] = vertex_bottom # 2*j + 1
            i_vertex += 1

            if j > 0:
                # 侧面三角形 1
                triangle = glm.uvec3(0, 0, 0)
                triangle[0] = 2*j + 1
                triangle[1] = 2*j
                triangle[2] = 2*j - 2
                indices[i_index] = triangle
                i_index += 1
                self.generate_temp_TBN(vertices[triangle[0]], vertices[triangle[1]], vertices[triangle[2]])

                # 侧面三角形 2
                triangle = glm.uvec3(0, 0, 0)
                triangle[0] = 2*j + 1
                triangle[1] = 2*j - 2
                triangle[2] = 2*j - 1
                indices[i_index] = triangle
                i_index += 1
                self.generate_temp_TBN(vertices[triangle[0]], vertices[triangle[1]], vertices[triangle[2]])

                yield

        del vertices[i_vertex:]
        del indices[i_index:]

    @property
    def n_divide(self):
        return self.__n_divide
    
    @n_divide.setter
    @Mesh.param_setter
    def n_divide(self, n_divide:int):
        self.__n_divide = n_divide

    @property
    def start_angle(self):
        return self.__start_angle
    
    @start_angle.setter
    @Mesh.param_setter
    def start_angle(self, angle:float):
        self.__start_angle = angle

    @property
    def span_angle(self):
        return self.__span_angle
    
    @span_angle.setter
    @Mesh.param_setter
    def span_angle(self, angle:float):
        self.__span_angle = angle

    @property
    def radius(self):
        return self.__radius
    
    @radius.setter
    @Mesh.param_setter
    def radius(self, radius:float):
        self.__radius = radius

    @property
    def height(self):
        return self.__height
    
    @height.setter
    @Mesh.param_setter
    def height(self, height:float):
        self.__height = height



from ..Mesh import Mesh

from glass.utils import checktype
from glass import Vertex

import glm
import math

class TorusFace(Mesh):

    @checktype
    def __init__(self, inner_radius:float=1, outer_radius:float=2,
                 color:(glm.vec3,glm.vec4)=glm.vec4(0.396, 0.74151, 0.69102, 1), back_color:(glm.vec3,glm.vec4)=None,
                 n_divide:int=100, start_angle:float=0, span_angle:float=360,
                 vertical:bool=False, normalize_tex_coord:bool=False,
                 name:str="", block:bool=True):
        Mesh.__init__(self, color=color, back_color=back_color, name=name, block=block)
        self.__inner_radius = inner_radius
        self.__outer_radius = outer_radius
        self.__start_angle = start_angle
        self.__span_angle = span_angle
        self.__n_divide = n_divide
        self.__vertical = vertical
        self.__normalize_tex_coord = normalize_tex_coord
        self.start_building()

    def build(self):
        self.is_closed = False
        self.self_calculated_normal = True

        vertices = self.vertices
        indices = self.indices
        inner_radius = self.__inner_radius
        outer_radius = self.__outer_radius
        start_angle = self.__start_angle/180*math.pi
        span_angle = self.__span_angle/180*math.pi
        n_divide = self.__n_divide
        vertical = self.__vertical
        normalize_tex_coord = self.__normalize_tex_coord

        i_vertex = 0
        i_index = 0

        tex_coord_outer_radius = outer_radius
        tex_coord_inner_radius = inner_radius
        if normalize_tex_coord:
            tex_coord_outer_radius = 0.5
            tex_coord_inner_radius = 0.5*inner_radius/outer_radius
        
        normal = glm.vec3(0, 0, 1)
        if vertical:
            normal = glm.vec3(0, -1, 0)

        for j in range(n_divide):
            theta = start_angle + span_angle*j/(n_divide-1)
            cos_theta = math.cos(theta)
            sin_theta = math.sin(theta)
            
            edge_inner = inner_radius * glm.vec3(cos_theta, sin_theta, 0)
            edge_outer = outer_radius * glm.vec3(cos_theta, sin_theta, 0)
            if vertical:
                edge_inner = inner_radius * glm.vec3(cos_theta, 0, sin_theta)
                edge_outer = outer_radius * glm.vec3(cos_theta, 0, sin_theta)
            
            vertex_inner = Vertex()
            vertex_inner.position = edge_inner
            vertex_inner.normal = normal
            vertex_inner.tex_coord = glm.vec3(0.5+tex_coord_inner_radius*cos_theta, 0.5+tex_coord_inner_radius*sin_theta, 0)

            vertex_outer = Vertex()
            vertex_outer.position = edge_outer
            vertex_outer.normal = normal
            vertex_outer.tex_coord = glm.vec3(0.5+tex_coord_outer_radius*cos_theta, 0.5+tex_coord_outer_radius*sin_theta, 0)

            vertices[i_vertex] = vertex_inner # 2*j
            i_vertex += 1

            vertices[i_vertex] = vertex_outer # 2*j + 1
            i_vertex += 1

            if j > 0:
                triangle = glm.uvec3(0, 0, 0)
                triangle[0] = 2*j + 1
                triangle[1] = 2*j
                triangle[2] = 2*j - 2
                indices[i_index] = triangle
                i_index += 1
                self.generate_temp_TBN(vertices[triangle[0]], vertices[triangle[1]], vertices[triangle[2]])

                triangle = glm.uvec3(0, 0, 0)
                triangle[0] = 2*j + 1
                triangle[1] = 2*j - 2
                triangle[2] = 2*j - 1
                indices[i_index] = triangle
                i_index += 1
                self.generate_temp_TBN(vertices[triangle[0]], vertices[triangle[1]], vertices[triangle[2]])

                yield
        
        del vertices[i_vertex:]
        del indices[i_index:]

    @property
    def start_angle(self):
        return self.__start_angle
    
    @start_angle.setter
    @Mesh.param_setter
    def start_angle(self, angle:float):
        self.__start_angle = angle

    @property
    def span_angle(self):
        return self.__span_angle
    
    @span_angle.setter
    @Mesh.param_setter
    def span_angle(self, angle:float):
        self.__span_angle = angle

    @property
    def n_divide(self):
        return self.__n_divide
    
    @n_divide.setter
    @Mesh.param_setter
    def n_divide(self, n_divide):
        self.__n_divide = n_divide

    @property
    def vertical(self):
        return self.__vertical
    
    @vertical.setter
    @Mesh.param_setter
    def vertical(self, flag:bool):
        self.__vertical = flag

    @property
    def inner_radius(self):
        return self.__inner_radius
    
    @inner_radius.setter
    @Mesh.param_setter
    def inner_radius(self, inner_radius:float):
        self.__inner_radius = inner_radius

    @property
    def outer_radius(self):
        return self.__outer_radius
    
    @outer_radius.setter
    @Mesh.param_setter
    def outer_radius(self, outer_radius:float):
        self.__outer_radius = outer_radius

    @property
    def normalize_tex_coord(self):
        return self.__normalize_tex_coord
    
    @normalize_tex_coord.setter
    @Mesh.param_setter
    def normalize_tex_coord(self, flag:bool):
        self.__normalize_tex_coord = flag


from ..Mesh import Mesh

from glass import Vertex
from glass.utils import checktype

import glm
from OpenGL import GL

class Polyline(Mesh):
    
    def __init__(self, points:list=[], color:(glm.vec3,glm.vec4)=glm.vec4(0.396, 0.74151, 0.69102, 1),
                 line_width:int=2, loop:bool=False,
                 name:str="", block:bool=True):
        primitive_type = (GL.GL_LINE_STRIP if not loop else GL.GL_LINE_LOOP)
        Mesh.__init__(self, primitive_type=primitive_type, color=color, name=name, block=block, shared=False)
        self.render_hints.line_width = line_width
        self.__points = points
        self.start_building()

    def build(self):
        points = self.__points
        vertices = self.vertices

        length = 0
        len_points = len(points)
        for i in range(len_points):
            if i > 0:
                length += glm.length(points[i]-points[i-1])
            bitangent = None
            if i < len_points-1:
                bitangent = glm.normalize(points[i+1] - points[i])
            else:
                bitangent = glm.normalize(points[i] - points[i-1])
            vertices[i] = Vertex(position=points[i], bitangent=bitangent, tex_coord=glm.vec3(length, 0, 0))

        del vertices[len_points:]

    @property
    def points(self):
        return self.__points
    
    @points.setter
    def points(self, points):
        self.__points = points
        self.start_building()

    @property
    def line_width(self):
        return self.render_hints.line_width
    
    @line_width.setter
    @checktype
    def line_width(self, line_width:int):
        self.render_hints.line_width = line_width

    @property
    def loop(self):
        return (self.primitive_type == GL.GL_LINE_LOOP)
    
    @loop.setter
    @Mesh.param_setter
    def loop(self, flag:bool):
        self.primitive_type = (GL.GL_LINE_STRIP if not flag else GL.GL_LINE_LOOP)